# coding: utf-8

# ------------------------------------------------------------------------
# Ensembles héréditairement finis (I)
#
# Marc Lorenzi
# 5 novembre 2021
#
# Fonctions documentées dans HF01.ipynb
# ------------------------------------------------------------------------

import matplotlib.pyplot as plt

def ppow(a, n):
    p = 0
    for k in range(n):
        p = a ** p
    return p

def rang(A):
    k = 0
    while A != []:
        A = A[-1]
        k = k + 1
    return k

# ------------------------------------------------------------------------
# Bijection de V sur N et sa réciproque

def phi(A):
    n = 0
    for x in A:
        n = n + 2 ** phi(x)
    return n

def psi(n):
    s = []
    m = 0
    while n != 0:
        if n % 2 == 1: s.append(psi(m))
        n = n // 2
        m = m + 1
    return s

def V(n):
    return psi(ppow(2, n + 1) - 1)

# ------------------------------------------------------------------------
# Ordre strict sur V

def inferieur_strict(A, B):
    if A == []: return B != []
    elif B == []: return False
    else:
        return inferieur_strict(A[-1], B[-1]) or (A[-1] == B[-1] and inferieur_strict(A[:-1], B[:-1]))

# ------------------------------------------------------------------------
# Entiers de von Neumann, sigma, tau

def neumann(n):
    u = []
    for k in range(n):
        u = u + [u]
    return u

def nu(n):
    s = 0
    for k in range(n):
        s = 2 ** s + s
    return s

def sigma(n):
    s = []
    for k in range(n): s = [s]
    return s

def tau(n):
    return [sigma(k) for k in range(n)]

def phi_tau(n):
    s = 0
    for k in range(1, n + 1):
        s = s + ppow(2, k)
    return s

# def tostr(A):
#     n = rang(A)
#     if A == neumann(n): return str(n) # 'ν' +
#     elif A == sigma(n): return 'σ' + str(n)
#     elif A == tau(n): return 'τ' + str(n)
#     else:
#         s = '{'
#         l = len(A)
#         for k in range(l):
#             s = s + tostr(A[k])
#             if k < l - 1: s = s + ','
#         s = s + '}'
#         return s

# ------------------------------------------------------------------------
# Arbre d'un ensemble H.F.

def tracer0(A, bornes):
    xmin, xmax, y, d = bornes
    xm = (xmin + xmax) / 2
    if A == []:
        plt.plot([xm], [y], 'or')
    else:
        n = len(A)
        dx = (xmax - xmin) / n
        xm = (xmin + xmax) / 2
        for k in range(n):
            x1 = xmin + k * dx
            x2 = xmin + (k + 1) * dx
            xm2 = (x1 + x2) / 2
            plt.plot([xm, xm2], [y, y - d], color=[0.5, 0.5, 0.5])
            tracer0(A[k], (x1, x2, y - d, d))
        plt.plot([xm], [y], 'og')

def tracer_arbre(A):
    plt.rcParams['figure.figsize'] = (14, 5)
    r = rang(A)
    tracer0(A, (0, 1, r, 1))
    plt.xticks([])
    plt.yticks(range(r + 1))
    plt.grid()

# ------------------------------------------------------------------------
# Successeur, prédécesseur

def succ(A):
    k = 0
    B = []
    while k < len(A) and A[k] == B:
        B = succ(B)
        k = k + 1
    return [B] + A[k:]

# def HF0(A, n):
#     for k in range(n):
#         yield A
#         if k < n: A = succ(A)

def pred(A):
    X = A[0]
    s = []
    while X != []:
        X = pred(X)
        s.append(X)
    s.reverse()
    return s + A[1:]

# ------------------------------------------------------------------------
# Iétrateur

def HF(A, n):
    if n >= 0:
        for k in range(n):
            yield A
            if k < n: A = succ(A)
    else:
        for k in range(-n):
            yield A
            if k < -n: A = pred(A)

# ------------------------------------------------------------------------
# Opérations sur les ensembles H.F.

def appartient(x, A):
    n = len(A)
    if n == 0: return False
    elif inferieur_strict(x, A[0]) or inferieur_strict(A[n - 1], x): return False
    else:
        i = 0
        j = n
        while j - i > 1:
            k = (i + j) // 2
            if x == A[k]: return True
            elif inferieur_strict(x, A[k]): j = k
            else: i = k
        return A[i] == x

def inclus(A, B):
    for x in A:
        if not appartient(x, B): return False
    return True

def intersection(A, B):
    C = []
    while A != [] and B != []:
        if inferieur_strict(A[0], B[0]):
            A = A[1:]
        elif inferieur_strict(B[0], A[0]):
            B = B[1:]
        else:
            C.append(A[0])
            A = A[1:]
            B = B[1:]
    return C

def reunion(A, B):
    C = []
    while A != [] and B != []:
        if inferieur_strict(A[0], B[0]):
            C.append(A[0])
            A = A[1:]
        elif inferieur_strict(B[0], A[0]):
            C.append(B[0])
            B = B[1:]
        else:
            C.append(A[0])
            A = A[1:]
            B = B[1:]
    C = C + B[:]
    C = C + A[:]
    return C

def difference(A, B):
    C = []
    while A != [] and B != []:
        if inferieur_strict(A[0], B[0]):
            C.append(A[0])
            A = A[1:]
        elif inferieur_strict(B[0], A[0]):
            B = B[1:]
        else:
            A = A[1:]
            B = B[1:]
    if B == []:
        C = C + A[:]
    return C

def parties(A):
    if A == []: return [[]]
    else:
        P = parties(A[1:])
        P1 = [reunion([A[0]], X) for X in P]
        return reunion(P, P1)

# ------------------------------------------------------------------------
# Couples

def couple(A, B):
    if A == B:
        return [[A]]
    elif inferieur_strict(A, B):
        return [[A], [A, B]]
    else:
        return [[A], [B, A]]

def composantes(C):
    A = C[0][0]
    if len(C) == 1:
        return (A, A)
    else:
        B = difference(C[1], C[0])[0]
        return (A, B)

def est_couple(A):
    if len(A) == 1 and len(A[0]) == 1: return True
    elif len(A) == 2 and len(A[0]) == 1 and len(A[1]) == 2 and inclus(A[0], A[1]): return True
    else: return False

# ------------------------------------------------------------------------
# Représentation formatée d'un ensemble H.F.

def tostring(A):
    n = rang(A)
    if A == neumann(n): return str(n) # 'ν' +
    elif A == sigma(n): return 'σ' + str(n)
    elif A == tau(n): return 'τ' + str(n)
    elif est_couple(A):
        x, y = composantes(A)
        return '<' + tostring(x) + ',' + tostring(y) + '>'
    else:
        s = '{'
        l = len(A)
        for k in range(l):
            s = s + tostring(A[k])
            if k < l - 1: s = s + ','
        s = s + '}'
        return s

# ------------------------------------------------------------------------
# Produit cartésien

def produit(A, B):
    C = []
    for x in A:
        for y in B:
            C = reunion(C, [couple(x, y)])
    return C
